/*
 *   khexedit - Versatile hex editor
 *   Copyright (C) 1999  Espen Sand, espensa@online.no
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <qframe.h>
#include <qlayout.h>
#include <qmsgbox.h> 
#include <qpushbutton.h> 

#include <kapp.h>
#include <kbuttonbox.h>

#include "dialogmargin.h"
#include "stringdialog.h"


CStringDialog::CStringDialog( QWidget *parent, const char *name )
  : QDialog( parent, name, false, WStyle_Customize|WStyle_DialogBorder ),
    mBusy( false )
{
  QString text;
  QVBoxLayout *topLayout = new QVBoxLayout( this, OUTER_SEP, INNER_SEP );  
  if( topLayout == 0 ) { return; }

  setCaption( i18n("Extract Strings" ) );

  QGridLayout *gbox = new QGridLayout( 3, 3, INNER_SEP );
  if( gbox == 0 ) { return; }
  topLayout->addLayout( gbox );
  gbox->setColStretch( 1, 10 );
 
  mLengthSpin = new QSpinBox( this );
  if( mLengthSpin == 0 ) { return; }
  mLengthSpin->setFixedHeight( mLengthSpin->sizeHint().height() );
  mLengthSpin->setMinimumWidth( fontMetrics().width("M")*20 );
  mLengthSpin->setRange( 1, 1000000 );
  mLengthSpin->setValue( 4 );
  gbox->addMultiCellWidget( mLengthSpin, 0, 0, 1, 2 );

  text = i18n("&Minimum length");
  QLabel *label = new QLabel( mLengthSpin, text, this );
  label->setFixedHeight( mLengthSpin->sizeHint().height() );
  label->setFixedWidth( label->sizeHint().width() );
  gbox->addWidget( label, 0, 0, AlignRight );

  mFilter = new QLineEdit( this );
  if( mFilter == 0 ) { return; }
  mFilter->setFixedHeight( mFilter->sizeHint().height() );
  mFilter->setMinimumWidth( fontMetrics().width("M")*20 );
  mFilter->setText("*");
  gbox->addWidget( mFilter, 1, 1 );
  
  text = i18n("&Filter");
  label = new QLabel( mFilter, text, this );
  label->setFixedHeight( mFilter->sizeHint().height() );
  label->setFixedWidth( label->sizeHint().width() );
  gbox->addWidget( label, 1, 0, AlignRight );

  QPushButton *useButton = new QPushButton( this, "use" );
  if( useButton == 0 ) { return; }
  useButton->setText( i18n("&Use") );
  useButton->setFixedHeight( mFilter->sizeHint().height() );
  useButton->setFixedWidth( useButton->sizeHint().width() );
  connect( useButton, SIGNAL(clicked()),this,SLOT(startCapture()) );
  gbox->addWidget( useButton, 1, 2 );

  text = i18n("&Ignore case");
  mCheckIngoreCase = new QCheckBox( text, this );
  mCheckIngoreCase->setMinimumSize( mCheckIngoreCase->sizeHint() );
  gbox->addWidget( mCheckIngoreCase, 2, 1, AlignLeft );

  mStringList = new QListBox ( this, "list" );
  mStringList->setFont( kapp->fixedFont ); 
  mStringList->setMinimumHeight( fontMetrics().height() * 20 );
  connect( mStringList, SIGNAL(highlighted(int)),
	   SLOT(selectionChanged(int)));
  topLayout->addWidget( mStringList, 10 );
  
  QHBoxLayout *hbox = new QHBoxLayout( INNER_SEP );
  if( hbox == 0 ) { return; }
  topLayout->addLayout( hbox );
  
  text = i18n("Number of strings:");
  label = new QLabel( text, this );
  label->setFixedHeight( label->sizeHint().height() );
  label->setFixedWidth( label->sizeHint().width() );
  hbox->addWidget( label, AlignLeft );
  
  mListSizeLabel = new QLabel( QString(), this );
  mListSizeLabel->setFixedWidth( fontMetrics().width("8")*10 );
  hbox->addWidget( mListSizeLabel, AlignLeft|AlignHCenter );

  text = i18n("Displayed:");
  label = new QLabel( text, this );
  label->setFixedHeight( label->sizeHint().height() );
  label->setFixedWidth( label->sizeHint().width() );
  hbox->addWidget( label, AlignLeft );
  
  mDisplaySizeLabel = new QLabel( QString(""), this );
  mDisplaySizeLabel->setFixedWidth( fontMetrics().width("8")*10 );
  hbox->addWidget( mDisplaySizeLabel, AlignLeft|AlignHCenter);
  
  hbox->addStretch ( 10 );  

  QFrame *hline = new QFrame( this );
  hline->setFrameStyle( QFrame::Sunken | QFrame::HLine );
  hline->setFixedHeight( hline->sizeHint().height() );
  topLayout->addWidget( hline );


  topLayout->addSpacing( ACTION_SPACE );

  KButtonBox *buttonBox = new KButtonBox( this, KButtonBox::HORIZONTAL, 0, 
					  INNER_SEP );

  QPushButton *helpButton = buttonBox->addButton( i18n("&Help"), false );
  buttonBox->addStretch();
  QPushButton *startButton = buttonBox->addButton( i18n("&Start"), false );
  QPushButton *cancelButton = buttonBox->addButton( i18n("&Cancel"), false );
  buttonBox->layout();
  startButton->setDefault( true );
  connect( cancelButton, SIGNAL(clicked()),this,SLOT(reject()) );
  connect( helpButton, SIGNAL(clicked()),this,SLOT(help()) );
  connect( startButton, SIGNAL(clicked()),this,SLOT(startCapture()) );
  topLayout->addWidget( buttonBox );

  topLayout->activate();
  setGeometry( x(), y(), minimumSize().width(), minimumSize().height() );

  updateListInfo();
}


CStringDialog::~CStringDialog( void )
{
}


void CStringDialog::help( void )
{
  kapp->invokeHTMLHelp( QString("khexedit/khexedit.html"), QString() );
}


void CStringDialog::startCapture( void )
{
  if( mBusy == true )
  {
    return;
  }


  mRegExp.setCaseSensitive( mCheckIngoreCase->isChecked()==true ? false:true );
  mRegExp.setWildcard( true );
  QString str = mFilter->text();
  if( str.isEmpty() == true )
  {
    str = "*";
  }
  mRegExp = str.data();
  if( mRegExp.isValid() == false )
  {
    QString msg = i18n(""
      "The filter expression you have specified is illegal.\n"
      "You must specify a valid regular expression.\n\n"
      "Proceed with no filter?");

    int reply = QMessageBox::warning( 
      this,
      i18n( "Extract strings" ), 
      msg,
      i18n( "&Yes" ),
      i18n( "&No" ),
      0, 0 );
    if( reply == 1 ) // No
    {
      return;
    }
    mRegExp = "*";
  }


  mStringData.minLength = mLengthSpin->value();

  removeList();

  mBusy = true;
  emit collect();
  mBusy = false;
}


void CStringDialog::startGoto( int index )
{
  const char *listText = mStringList->text( index );
  if( listText == 0 )
  {
    return;
  }
  uint offset,val1,val2;
  sscanf( listText, "%X:%X", &val1, &val2 );
  offset = (val1<<16) + val2;

  uint size = strlen( listText ); 
  size = size > 13 ? size - 13 : 1; 

  emit markText( offset, size, true );
}


void CStringDialog::selectionChanged( int index )
{
  killTimers();
  startTimer( 200 );
}



void CStringDialog::timerEvent( QTimerEvent *e )
{
  killTimers();

  int index = mStringList->currentItem();
  if( index == -1 )
  {
    return;
  }
  startGoto( index );
}


int CStringDialog::setList( QList<QString> &list, CProgress &p )
{
  clearList();
  enableList( false );

  for( uint i=0; i < list.count(); i++ )
  {
    const QString *str = list.at(i);
    if( str != 0 )
    {
      appendListItem( *str );
    }

    if( (i+1) % 200 == 0 )
    {
      updateListInfo();

      int errCode = p.step( (float)i/(float)list.count() );
      if( errCode == Err_Stop )
      {
	p.finish();
	enableList( true );
	return( errCode );
      }
    }
  }

  p.finish();
  updateListInfo();
  enableList( true );

  return( Err_Success );
}


void CStringDialog::clearList( void )
{
  if( mStringList->count() > 0 )
  {
    mStringList->clear();
    mStringList->update();
  }
  updateListInfo();
}


void CStringDialog::removeList( void )
{
  mStringData.clear();
  clearList();
}


void CStringDialog::enableList( bool state )
{
  mStringList->setAutoUpdate( state );
  mStringList->setEnabled( state );
  if( state == true )
  {
    mStringList->update();
  }
}


void CStringDialog::appendListItem( const QString &str )
{
  if( mRegExp.match( str.data(), 13 ) != -1 )
  {
    mStringList->insertItem( str.data() );
  }
}


void CStringDialog::updateListInfo( void )
{
  QString msg;
  msg.sprintf( "%u", mStringData.count() );
  mListSizeLabel->setText( msg );
  msg.sprintf( "%u", mStringList->count() );
  mDisplaySizeLabel->setText( msg );
}



